/*
Nintendulator - A Win32 NES emulator written in C.
Designed for maximum emulation accuracy.
Copyright (c) 2002  Quietust

Based on NinthStar, a portable Win32 NES Emulator written in C++
Copyright (C) 2000  David de Regt

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

For a copy of the GNU General Public License, go to:
http://www.gnu.org/copyleft/gpl.html#SEC1
*/

#include "Nintendulator.h"
#include "NES.h"
#include "CPU.h"
#include "PPU.h"
#include "APU.h"
#include "GFX.h"
#include "Debugger.h"
#include "States.h"
#include "Controllers.h"

struct tNES NES;
unsigned char PRG_ROM[0x800][0x1000];	// 8192 KB
unsigned char CHR_ROM[0x1000][0x400];	// 4096 KB
unsigned char PRG_RAM[0x10][0x1000];	//   64 KB
unsigned char CHR_RAM[0x20][0x400];	//   32 KB

void	NES_Init (void)
{
	PPU_Init();
	APU_Init();
	GFX_Init();
	Debugger_Init();
	States_Init();
	Controllers_Init();
	MapperInterface_Init();
	Debugger_SetMode(0);
	NES_LoadSettings();
	NES_CloseFile();

	NES.Stopped = TRUE;
	NES.Stop = FALSE;
	NES.NeedQuit = FALSE;
	NES.GameGenie = FALSE;
	NES.ROMLoaded = FALSE;

	GFX_UpdateTitlebar();
}

void	NES_Release (void)
{
	if (NES.ROMLoaded)
		NES_CloseFile();
	APU_Release();
	GFX_Release();
	Controllers_Release();
	MapperInterface_Release();

	NES_SaveSettings();
	DestroyWindow(mWnd);
}

void	NES_OpenFile (char *filename)
{
	int len = strlen(filename);
	BOOL LoadRet;
	if (NES.ROMLoaded)
		NES_CloseFile();

	NES.IsNSF = FALSE;
	if (!stricmp(filename + len - 4, ".NES"))
		LoadRet = NES_OpenFileiNES(filename);
	else if (!stricmp(filename + len - 4, ".NSF"))
		LoadRet = NES_OpenFileNSF(filename);
	else if (!stricmp(filename + len - 4, ".UNF"))
		LoadRet = NES_OpenFileUNIF(filename);
	else if (!stricmp(filename + len - 5, ".UNIF"))
		LoadRet = NES_OpenFileUNIF(filename);
	else if (!stricmp(filename + len - 4, ".FDS"))
		LoadRet = NES_OpenFileFDS(filename);
	else if (!stricmp(filename + len - 4, ".NRF"))
		LoadRet = NES_OpenFileNRFF(filename);
	else if (!stricmp(filename + len - 5, ".NRFF"))
		LoadRet = NES_OpenFileNRFF(filename);
	else	LoadRet = FALSE;

	if (!LoadRet)
	{
		MessageBox(mWnd,"File type not recognized or mapper not supported!","Nintendulator",MB_OK | MB_ICONERROR);
		return;
	}
	NES.ROMLoaded = TRUE;
	States_SetFilename(filename);

	Debugger.NTabChanged = TRUE;
	Debugger.PalChanged = TRUE;
	Debugger.PatChanged = TRUE;

	NES_Reset(0);
}

void	NES_CloseFile (void)
{
	int i;
	if ((MI) && (MI->UnloadMapper))
		MI->UnloadMapper();
	MI = NULL;
	NES.ROMLoaded = FALSE;

	for (i = 0; i < 16; i++)
	{
		CPU.PRGPointer[i] = PRG_RAM[0];
		CPU.ReadHandler[i] = CPU_ReadPRG;
		CPU.WriteHandler[i] = CPU_WritePRG;
		CPU.Readable[i] = FALSE;
		CPU.Writable[i] = FALSE;
		PPU.CHRPointer[i] = CHR_RAM[0];
	}
	for (i = 0; i < 12; i++)
		PPU.Writable[i] = FALSE;
}

BOOL	NES_OpenFileiNES (char *filename)
{
	FILE *in;
	unsigned char HeaderBs[4], Trainer[0x200];
	unsigned long tmp;
	BOOL p2Found = FALSE;
	unsigned char MapperNum;

	in = fopen(filename,"rb");
	fread(&tmp,4,1,in);
	if (tmp != 0x1A53454E)
	{
		fclose(in);
		return FALSE;
	}
	fread(HeaderBs,1,4,in);
	fseek(in,8,SEEK_CUR);
	
	MapperNum = ((HeaderBs[2] & 0xF0) >> 4) | (HeaderBs[3] & 0xF0);
	MP.Flags = HeaderBs[2] | (HeaderBs[3] << 8);
	if (HeaderBs[2] & 0x04)
		fread(Trainer,1,0x200,in);

	MP.PRG_ROM_Size = HeaderBs[0] * 0x4000;
	fread(PRG_ROM,1,MP.PRG_ROM_Size,in);
	MP.CHR_ROM_Size = HeaderBs[1] * 0x2000;
	fread(CHR_ROM,1,MP.CHR_ROM_Size,in);

	fclose(in);

	NES.PRGMask = (HeaderBs[0] << 2) - 1;
	NES.CHRMask = (HeaderBs[1] << 3) - 1;

	for (tmp = 1; tmp < 0x40000000; tmp <<= 1)
		if (tmp == HeaderBs[0])
			p2Found = TRUE;
	if (!p2Found)
		NES.PRGMask = 0xFFFFFFFF;

	return MapperInterface_LoadByNum(MapperNum);
}

BOOL	NES_OpenFileUNIF (char *filename)
{
	char UNIFSig[4], BlockSig[4];
	unsigned long BlockHeader[2];
	char BoardName[64];
	unsigned char *tPRG[0x10], *tCHR[0x10];
	int PRGSize[0x10], CHRSize[0x10];
	int i;
	unsigned char tp;
	unsigned char *PRGPoint, *CHRPoint;
	BOOL p2Found;
	int p2;

	FILE *in = fopen(filename,"rb");
	if (!in)
		return FALSE;
	fread(UNIFSig,1,4,in);
	if (strncmp(UNIFSig,"UNIF",4))
	{
		fclose(in);
		return FALSE;
	}

	fseek(in,28,SEEK_CUR);

	for (i=0;i<0x10;i++)
	{	tPRG[i] = 0; tCHR[i] = 0; }

	MP.Flags = 0;
	while (!feof(in))
	{
		fread(BlockSig,1,4,in);
		fread(&BlockHeader[1],4,1,in);
		if (feof(in))
			continue;
		
		if (!strncmp(BlockSig,"MAPR",4)) {
			fread(BoardName,1,BlockHeader[1],in);
		} else if (!strncmp(BlockSig,"TVCI",4)) {
			fread(&tp,1,1,in);
			if (tp == 0) NES_SetCPUMode(0);
			if (tp == 1) NES_SetCPUMode(1);
			if (tp == 2) NES_SetCPUMode(0);
		} else if (!strncmp(BlockSig,"BATR",4)) {
			fread(&tp,1,1,in);
			MP.Flags |= 2;
		} else if (!strncmp(BlockSig,"MIRR",4)) {
			fread(&tp,1,1,in);
			MP.Flags |= (tp & 0x0F) << 8;
		} else if (!strncmp(BlockSig,"PRG0",4)) { PRGSize[0x0] = BlockHeader[1]; tPRG[0x0] = (unsigned char *) malloc(BlockHeader[1]); fread(tPRG[0x0], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"PRG1",4)) { PRGSize[0x1] = BlockHeader[1]; tPRG[0x1] = (unsigned char *) malloc(BlockHeader[1]); fread(tPRG[0x1], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"PRG2",4)) { PRGSize[0x2] = BlockHeader[1]; tPRG[0x2] = (unsigned char *) malloc(BlockHeader[1]); fread(tPRG[0x2], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"PRG3",4)) { PRGSize[0x3] = BlockHeader[1]; tPRG[0x3] = (unsigned char *) malloc(BlockHeader[1]); fread(tPRG[0x3], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"PRG4",4)) { PRGSize[0x4] = BlockHeader[1]; tPRG[0x4] = (unsigned char *) malloc(BlockHeader[1]); fread(tPRG[0x4], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"PRG5",4)) { PRGSize[0x5] = BlockHeader[1]; tPRG[0x5] = (unsigned char *) malloc(BlockHeader[1]); fread(tPRG[0x5], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"PRG6",4)) { PRGSize[0x6] = BlockHeader[1]; tPRG[0x6] = (unsigned char *) malloc(BlockHeader[1]); fread(tPRG[0x6], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"PRG7",4)) { PRGSize[0x7] = BlockHeader[1]; tPRG[0x7] = (unsigned char *) malloc(BlockHeader[1]); fread(tPRG[0x7], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"PRG8",4)) { PRGSize[0x8] = BlockHeader[1]; tPRG[0x8] = (unsigned char *) malloc(BlockHeader[1]); fread(tPRG[0x8], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"PRG9",4)) { PRGSize[0x9] = BlockHeader[1]; tPRG[0x9] = (unsigned char *) malloc(BlockHeader[1]); fread(tPRG[0x9], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"PRGA",4)) { PRGSize[0xA] = BlockHeader[1]; tPRG[0xA] = (unsigned char *) malloc(BlockHeader[1]); fread(tPRG[0xA], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"PRGB",4)) { PRGSize[0xB] = BlockHeader[1]; tPRG[0xB] = (unsigned char *) malloc(BlockHeader[1]); fread(tPRG[0xB], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"PRGC",4)) { PRGSize[0xC] = BlockHeader[1]; tPRG[0xC] = (unsigned char *) malloc(BlockHeader[1]); fread(tPRG[0xC], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"PRGD",4)) { PRGSize[0xD] = BlockHeader[1]; tPRG[0xD] = (unsigned char *) malloc(BlockHeader[1]); fread(tPRG[0xD], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"PRGE",4)) { PRGSize[0xE] = BlockHeader[1]; tPRG[0xE] = (unsigned char *) malloc(BlockHeader[1]); fread(tPRG[0xE], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"PRGF",4)) { PRGSize[0xF] = BlockHeader[1]; tPRG[0xF] = (unsigned char *) malloc(BlockHeader[1]); fread(tPRG[0xF], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"CHR0",4)) { CHRSize[0x0] = BlockHeader[1]; tCHR[0x0] = (unsigned char *) malloc(BlockHeader[1]); fread(tCHR[0x0], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"CHR1",4)) { CHRSize[0x1] = BlockHeader[1]; tCHR[0x1] = (unsigned char *) malloc(BlockHeader[1]); fread(tCHR[0x1], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"CHR2",4)) { CHRSize[0x2] = BlockHeader[1]; tCHR[0x2] = (unsigned char *) malloc(BlockHeader[1]); fread(tCHR[0x2], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"CHR3",4)) { CHRSize[0x3] = BlockHeader[1]; tCHR[0x3] = (unsigned char *) malloc(BlockHeader[1]); fread(tCHR[0x3], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"CHR4",4)) { CHRSize[0x4] = BlockHeader[1]; tCHR[0x4] = (unsigned char *) malloc(BlockHeader[1]); fread(tCHR[0x4], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"CHR5",4)) { CHRSize[0x5] = BlockHeader[1]; tCHR[0x5] = (unsigned char *) malloc(BlockHeader[1]); fread(tCHR[0x5], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"CHR6",4)) { CHRSize[0x6] = BlockHeader[1]; tCHR[0x6] = (unsigned char *) malloc(BlockHeader[1]); fread(tCHR[0x6], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"CHR7",4)) { CHRSize[0x7] = BlockHeader[1]; tCHR[0x7] = (unsigned char *) malloc(BlockHeader[1]); fread(tCHR[0x7], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"CHR8",4)) { CHRSize[0x8] = BlockHeader[1]; tCHR[0x8] = (unsigned char *) malloc(BlockHeader[1]); fread(tCHR[0x8], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"CHR9",4)) { CHRSize[0x9] = BlockHeader[1]; tCHR[0x9] = (unsigned char *) malloc(BlockHeader[1]); fread(tCHR[0x9], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"CHRA",4)) { CHRSize[0xA] = BlockHeader[1]; tCHR[0xA] = (unsigned char *) malloc(BlockHeader[1]); fread(tCHR[0xA], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"CHRB",4)) { CHRSize[0xB] = BlockHeader[1]; tCHR[0xB] = (unsigned char *) malloc(BlockHeader[1]); fread(tCHR[0xB], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"CHRC",4)) { CHRSize[0xC] = BlockHeader[1]; tCHR[0xC] = (unsigned char *) malloc(BlockHeader[1]); fread(tCHR[0xC], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"CHRD",4)) { CHRSize[0xD] = BlockHeader[1]; tCHR[0xD] = (unsigned char *) malloc(BlockHeader[1]); fread(tCHR[0xD], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"CHRE",4)) { CHRSize[0xE] = BlockHeader[1]; tCHR[0xE] = (unsigned char *) malloc(BlockHeader[1]); fread(tCHR[0xE], 1, BlockHeader[1], in);
		} else if (!strncmp(BlockSig,"CHRF",4)) { CHRSize[0xF] = BlockHeader[1]; tCHR[0xF] = (unsigned char *) malloc(BlockHeader[1]); fread(tCHR[0xF], 1, BlockHeader[1], in);
		} else {
			fseek(in,BlockHeader[1],SEEK_CUR);
		}
	}

	PRGPoint = PRG_ROM[0];
	CHRPoint = CHR_ROM[0];
	
	for (i = 0; i < 0x10; i++)
	{
		if (tPRG[i])
		{
			memcpy(PRGPoint, tPRG[i], PRGSize[i]);
			PRGPoint += PRGSize[i];
			free(tPRG[i]);
		}
		if (tCHR[i])
		{
			memcpy(CHRPoint, tCHR[i], CHRSize[i]);
			CHRPoint += CHRSize[i];
			free(tCHR[i]);
		}
	}

	fclose(in);

	MP.PRG_ROM_Size = PRGPoint - PRG_ROM[0];
	MP.CHR_ROM_Size = CHRPoint - CHR_ROM[0];
	NES.PRGMask = (MP.PRG_ROM_Size / 0x1000) - 1;
	NES.CHRMask = (MP.CHR_ROM_Size / 0x400) - 1;
	p2Found = FALSE;
	for (p2 = 1; p2 < 0x40000000; p2 <<= 1)
		if (p2 == MP.PRG_ROM_Size)
			p2Found = TRUE;
	if (!p2Found) NES.PRGMask = 0xffffffff;

	return	MapperInterface_LoadByBoard(BoardName);
}

BOOL	NES_OpenFileFDS (char *filename)
{
	FILE *in;
	unsigned char Header[16];
	int i;

	in = fopen(filename,"rb");
	fread(Header,1,16,in);
	if (strncmp(Header,"FDS\x1a",4))
	{
		fclose(in);
		return FALSE;
	}
	for (i = 0; i < Header[4]; i++)
		fread(PRG_ROM[i << 4],1,65500,in);
	fclose(in);

	MP.PRG_ROM_Size = 0x10000 * Header[4];
	NES.PRGMask = (Header[4] << 4) - 1;

	if ((!MapperInterface_LoadByNum(20)) || (!MapperInterface_LoadByBoard("HVC-FMR")))
	{
		MessageBox(mWnd,"FDS mapper support not found!","Error",MB_OK);
		return FALSE;
	}
	return TRUE;
}

BOOL	NES_OpenFileNSF (char *filename)
{
	FILE *in;
	unsigned char Header[128];		//Header Bytes
	int ROMlen;

	in = fopen(filename,"rb");
	fseek(in,0,SEEK_END);
	ROMlen = ftell(in) - 128;
	fseek(in,0,SEEK_SET);
	fread(Header,1,128,in);
	if (strncmp((char *)&Header[0],"NESM\x1a\x01",6))
		return FALSE;
	if (memcmp(&Header[0x70],"\0\0\0\0\0\0\0\0",8))
		fread(&PRG_ROM[0][0] + ((Header[0x8] | (Header[0x9] << 8)) & 0x0FFF),1,ROMlen,in);
	else
	{
		memcpy(&Header[0x70],"\x00\x01\x02\x03\x04\x05\x06\x07",8);
		fread(&PRG_ROM[0][0] + ((Header[0x8] | (Header[0x9] << 8)) & 0x7FFF),1,ROMlen,in);
	}
	fclose(in);

	MP.Flags = Header[0x7B];
	MP.PRG_ROM_Size = 0x800000;
	NES.PRGMask = 0x7FFFFF;
	{
		unsigned char NSFcode[139] = {0x47,0x50,0x16,0x50,0x8A,0x50,Header[0x70],Header[0x71],Header[0x72],Header[0x73],Header[0x74],Header[0x75],Header[0x76],Header[0x77],Header[7]-1,Header[6]-1,0x4C,Header[0xA],Header[0xB],0x4C,Header[0xC],Header[0xD],0x78,0xD8,0xA2,0xFF,0x9A,0xAD,0x0E,0x50,0x48,0xA9,0x00,0x48,0xA2,0xFD,0x9A,0xA2,0x07,0xBD,0x06,0x50,0x9D,0xF8,0x5F,0xCA,0x10,0xF7,0xE8,0xA9,0x0F,0x8D,0x15,0x40,0xA9,0x80,0x8D,0x00,0x20,0x8E,0x01,0x20,0xAD,0xFF,0x01,0x20,0x10,0x50,0x4C,0x44,0x50,0x20,0x13,0x50,0xB8,0xA0,0x01,0x8C,0x16,0x40,0x88,0x8C,0x16,0x40,0xA2,0x08,0xAD,0x16,0x40,0x4A,0x98,0x2A,0xA8,0xCA,0xD0,0xF6,0x4D,0xFE,0x01,0x8C,0xFE,0x01,0x2D,0xFE,0x01,0x4A,0x90,0x0D,0xAD,0xFF,0x01,0xCD,0x0F,0x50,0xF0,0xAE,0xEE,0xFF,0x01,0x50,0xA9,0x4A,0x90,0x0A,0xAD,0xFF,0x01,0xF0,0xA1,0xCE,0xFF,0x01,0x50,0x9C,0x29,0x04,0xD0,0x98,0x40};
		memcpy(PRG_RAM[2],NSFcode,sizeof(NSFcode));
	}

	if ((!MapperInterface_LoadByNum(666)) || (!MapperInterface_LoadByBoard("NSF")))
	{
		MessageBox(mWnd,"NSF mapper support not found!","Error",MB_OK);
		return FALSE;
	}

	NES.IsNSF = TRUE;
	return TRUE;
}
int	__cdecl NSFReader (int Bank, int Addy)
{
	if ((Bank == 0xF) && (Addy >= 0xFFA))
		return CPU_ReadPRG(0x5,Addy - 0xFFA);
	else	return CPU_ReadPRG(Bank,Addy);
}

//#define NRFF_HEADER	0x4646524E
#define NRFF_HEADER	'FFRN'
//#define NRFF_BORD	0x44524F42
#define NRFF_BORD	'DROB'
//#define NRFF_TVCI	0x49435654
#define NRFF_TVCI	'ICVT'
//#define NRFF_BATT	0x54544142
#define NRFF_BATT	'TTAB'
//#define NRFF_MIRR	0x5252494D
#define NRFF_MIRR	'RRIM'
//#define NRFF_PRGM	0x4D475250
#define NRFF_PRGM	'MGRP'
//#define NRFF_CHAR	0x52414843
#define NRFF_CHAR	'RAHC'

BOOL	NES_OpenFileNRFF (char *filename)
{
	unsigned long BlockSig, tp, i;
	int BlockHeader;
	char *BoardName;
	unsigned char *tbuf, *PRGPoint = PRG_ROM[0], *CHRPoint = CHR_ROM[0];
	BOOL p2Found = FALSE;
	int p2;
	FILE *in = fopen(filename,"rb");

	if (!in)
		return FALSE;
	fread(&BlockSig,1,4,in);
	if (BlockSig != NRFF_HEADER)
	{
		fclose(in);
		return FALSE;
	}
	fseek(in,4,SEEK_CUR);
	MP.Flags = 0;
	while (!feof(in))
	{
		fread(&BlockSig,1,4,in);
		fread(&BlockHeader,1,4,in);
		if (feof(in))
			break;
		switch (BlockSig)
		{
		case NRFF_BORD:	BoardName = (char *)malloc(BlockHeader);
				fread(BoardName,1,BlockHeader,in);	break;
		case NRFF_TVCI:	fread(&tp,1,4,in);
				if (BlockHeader-4)
					fseek(in,BlockHeader-4,SEEK_CUR);
				if (tp == 0) NES_SetCPUMode(0);
				if (tp == 1) NES_SetCPUMode(1);		break;
		case NRFF_BATT:	MP.Flags |= 0x02;			break;
		case NRFF_MIRR:	fread(&tp,1,4,in);
				if (BlockHeader-4)
					fseek(in,BlockHeader-4,SEEK_CUR);
				if (tp == 3)
					tp = 5;
				MP.Flags |= (tp & 0x0F) << 8;		break;
		case NRFF_PRGM:	fread(&tp,1,4,in);
				fseek(in,4,SEEK_CUR);
				BlockHeader -= 8;
				if (BlockHeader < 0)
				{
					MessageBox(mWnd,"PRGM block corrupt!","Error",MB_OK);
					fclose(in);
					return FALSE;
				}
				if (BlockHeader)
				{
					tbuf = (unsigned char *)malloc(BlockHeader);
					fread(tbuf, 1, BlockHeader, in);
					for (i = 0; i < tp; i += BlockHeader)
						memcpy(PRGPoint+i,tbuf,BlockHeader);
				}
				else	memset(PRGPoint,0,tp);	// open bus
				PRGPoint += tp;				break;
		case NRFF_CHAR:	fread(&tp,1,4,in);
				fseek(in,4,SEEK_CUR);
				BlockHeader -= 8;
				if (BlockHeader < 0)
				{
					MessageBox(mWnd,"CHAR block corrupt!","Error",MB_OK);
					return FALSE;
				}
				if (BlockHeader)
				{
					tbuf = (unsigned char *)malloc(BlockHeader);
					fread(tbuf, 1, BlockHeader, in);
					for (i = 0; i < tp; i += BlockHeader)
						memcpy(CHRPoint+i,tbuf,BlockHeader);
				}
				else	memset(CHRPoint,0,tp);	// open bus
				CHRPoint += tp;				break;
		default:	fseek(in,BlockHeader,SEEK_CUR);		break;
		}
	}
	fclose(in);
	MP.PRG_ROM_Size = PRGPoint - PRG_ROM[0];
	MP.CHR_ROM_Size = CHRPoint - CHR_ROM[0];
	NES.PRGMask = (MP.PRG_ROM_Size >> 12) - 1;
	NES.CHRMask = (MP.CHR_ROM_Size >> 10) - 1;
	for (p2 = 1; p2 < 0x10000000; p2 <<= 1)
		if (p2 == MP.PRG_ROM_Size) p2Found = FALSE;
	if (!p2Found) NES.PRGMask = 0xFFFFFFFF;
	p2Found = MapperInterface_LoadByBoard(BoardName);
	free(BoardName);
	return p2Found;
}

void	NES_SetCPUMode (int NewMode)
{
	HMENU hMenu = GetMenu(mWnd);
	if (NewMode == 0)
	{
		CheckMenuRadioItem(hMenu,ID_PPU_MODE_NTSC,ID_PPU_MODE_PAL,ID_PPU_MODE_NTSC,MF_BYCOMMAND);
		PPU.SLEndFrame = 262;
		GFX.WantFPS = 60;
		GFX_LoadPalette(0);
		APU_SetFPS(60);
		PPU.IsPAL = FALSE;
	}
	else
	{
		CheckMenuRadioItem(hMenu,ID_PPU_MODE_NTSC,ID_PPU_MODE_PAL,ID_PPU_MODE_PAL,MF_BYCOMMAND);
		PPU.SLEndFrame = 312;
		GFX.WantFPS = 50;
		GFX_LoadPalette(1);
		APU_SetFPS(50);
		PPU.IsPAL = TRUE;
	}
}

unsigned char GameGeniePRG[0x1000] = {
0x78,0xD8,0xA9,0x00,0x8D,0x00,0x20,0xA2,0xFF,0x9A,0xA9,0x00,0x8D,0xF0,0xFF,0x20,0x1E,0xF0,0x8D,0xF1,0xFF,0x20,0x1E,0xF0,0x8D,0xF0,0xFF,0x4C,0x29,0xF0,0xA2,0x60,
0xA0,0x08,0xCA,0xD0,0xFD,0x88,0xD0,0xFA,0x60,0xA2,0x0A,0xAD,0x02,0x20,0x10,0xFB,0xCA,0xD0,0xF8,0xA2,0xFF,0x9A,0xA9,0x07,0x85,0x01,0xA9,0x00,0x85,0x00,0xA8,0x91,
0x00,0xC8,0xD0,0xFB,0xC6,0x01,0x10,0xF7,0xA9,0x06,0x85,0x01,0x8D,0x01,0x20,0xA9,0x80,0x85,0x00,0x8D,0x00,0x20,0xCE,0x2B,0x00,0x4C,0x1F,0xF4,0xA9,0x01,0x8D,0x16,
0x40,0xA9,0x00,0x8D,0x16,0x40,0xA0,0x08,0xAD,0x16,0x40,0x6A,0x26,0x07,0x88,0xD0,0xF7,0xA0,0x08,0xAD,0x17,0x40,0x6A,0x26,0x08,0x88,0xD0,0xF7,0x60,0x48,0x8A,0x48,
0x98,0x48,0xA5,0x0C,0xD0,0x17,0x20,0x69,0xF1,0x20,0x74,0xF1,0xAD,0x45,0x00,0x8D,0x05,0x20,0xAD,0x46,0x00,0x8D,0x05,0x20,0xA5,0x00,0x8D,0x00,0x20,0xA9,0x01,0x8D,
0x49,0x00,0x68,0xA8,0x68,0xAA,0x68,0x40,0x86,0x17,0x84,0x18,0x20,0x38,0xF2,0xA0,0x00,0xB1,0x39,0x85,0x0D,0xC8,0xB1,0x39,0x85,0x0E,0xAD,0x39,0x00,0x18,0x69,0x02,
0x8D,0x39,0x00,0xAD,0x3A,0x00,0x69,0x00,0x8D,0x3A,0x00,0xA9,0x00,0x85,0x16,0xA9,0x00,0x85,0x15,0x20,0xE7,0xF0,0xE6,0x15,0xA5,0x15,0xC5,0x0D,0xD0,0xF5,0xE6,0x16,
0xA5,0x16,0xC5,0x0E,0xD0,0xE9,0x60,0xA5,0x16,0xA6,0x0D,0x20,0x5A,0xF1,0x18,0x65,0x15,0x8D,0x40,0x00,0x4A,0xA8,0xB1,0x39,0x85,0x04,0xAD,0x40,0x00,0x29,0x01,0xF0,
0x07,0xA5,0x04,0x29,0x0F,0x4C,0x0E,0xF1,0xA5,0x04,0x4A,0x4A,0x4A,0x4A,0x85,0x19,0xA5,0x17,0x18,0x65,0x15,0x38,0xE9,0x04,0x85,0x1E,0xA5,0x18,0x18,0x65,0x16,0x85,
0x1F,0xA5,0x1F,0x0A,0x0A,0x0A,0x85,0x20,0xA9,0x00,0x06,0x20,0x2A,0x06,0x20,0x2A,0x18,0x65,0x05,0x85,0x21,0xA5,0x20,0x18,0x65,0x1E,0x85,0x20,0xA5,0x21,0x69,0x00,
0x85,0x21,0xA5,0x21,0x8D,0x41,0x04,0xA5,0x20,0x8D,0x42,0x04,0xA5,0x19,0x8D,0x43,0x04,0xA9,0x01,0x8D,0x40,0x04,0x20,0xB9,0xF1,0x60,0x85,0x27,0xA9,0x00,0xE0,0x00,
0xF0,0x06,0x18,0x65,0x27,0xCA,0xD0,0xFB,0x60,0xA9,0x00,0x8D,0x03,0x20,0xA9,0x02,0x8D,0x14,0x40,0x60,0xA9,0x64,0x85,0x2C,0xA4,0x29,0xC4,0x2A,0xF0,0x38,0xB9,0x00,
0x03,0xAA,0x18,0x69,0x05,0x85,0x28,0xA5,0x2C,0x38,0xE5,0x28,0x90,0x28,0x85,0x2C,0xA5,0x2B,0x18,0x79,0x00,0x03,0x69,0x03,0x85,0x2B,0xC8,0xB9,0x00,0x03,0x8D,0x06,
0x20,0xC8,0xB9,0x00,0x03,0x8D,0x06,0x20,0xC8,0xB9,0x00,0x03,0x8D,0x07,0x20,0xC8,0xCA,0xD0,0xF6,0x4C,0x7A,0xF1,0x84,0x29,0x60,0xAD,0x40,0x04,0x18,0x69,0x03,0x85,
0x04,0xA5,0x2B,0xC5,0x04,0xB0,0x0A,0xA5,0x0C,0xF0,0xEE,0x20,0x74,0xF1,0x4C,0xB9,0xF1,0x38,0xE5,0x04,0x85,0x2B,0xA2,0x00,0xA4,0x2A,0xBD,0x40,0x04,0x99,0x00,0x03,
0xE8,0xC8,0xE4,0x04,0xD0,0xF4,0x84,0x2A,0x60,0x85,0x31,0x20,0x38,0xF2,0xA0,0x00,0xB1,0x39,0x85,0x34,0xAA,0xC8,0xB1,0x39,0x85,0x35,0x20,0x5A,0xF1,0x85,0x37,0x18,
0x69,0x02,0x20,0xB0,0xF5,0x85,0x38,0xAA,0xBD,0x63,0x05,0xAA,0xA5,0x34,0x9D,0x95,0x05,0xA5,0x35,0x9D,0x96,0x05,0xE8,0xE8,0xA0,0xFF,0xC8,0xB9,0x63,0x04,0xD0,0xFA,
0x98,0x9D,0x95,0x05,0xE8,0xC6,0x37,0xD0,0xF1,0xA6,0x38,0xA5,0x31,0x20,0xE0,0xF2,0xA6,0x38,0x20,0x5B,0xF2,0xA5,0x38,0x60,0x85,0x04,0xA9,0xBE,0x85,0x39,0xA9,0xFE,
0x85,0x3A,0xA5,0x04,0x10,0x02,0xE6,0x3A,0x0A,0xA8,0xB1,0x39,0x48,0xC8,0xB1,0x39,0x18,0x69,0xBE,0x85,0x39,0x68,0x69,0xFE,0x85,0x3A,0x60,0xBD,0x63,0x05,0xAA,0xBD,
0x95,0x05,0x85,0x34,0xBD,0x96,0x05,0x85,0x35,0xE8,0xE8,0xA5,0x2F,0x85,0x3C,0xA5,0x30,0x85,0x3D,0xA5,0x3D,0xD0,0x54,0xA5,0x34,0x85,0x3B,0xA5,0x2D,0x85,0x3E,0xA5,
0x2E,0x85,0x3F,0xA5,0x3F,0xD0,0x17,0xBD,0x95,0x05,0xA8,0xB9,0x23,0x05,0xF0,0x0E,0xA5,0x3E,0x99,0xA3,0x04,0xA5,0x3C,0x99,0xE3,0x04,0xE8,0x4C,0xA8,0xF2,0xBD,0x95,
0x05,0xA8,0xA9,0xFF,0x99,0xE3,0x04,0xE8,0xA5,0x3E,0x18,0x69,0x08,0x85,0x3E,0xA5,0x3F,0x69,0x00,0x85,0x3F,0xC6,0x3B,0xD0,0xCA,0xA5,0x3C,0x18,0x69,0x08,0x85,0x3C,
0xA5,0x3D,0x69,0x00,0x85,0x3D,0xC6,0x35,0xD0,0xA9,0x60,0xA5,0x34,0x85,0x3B,0xBD,0x95,0x05,0xA8,0xA9,0xFF,0x99,0xE3,0x04,0xE8,0xC6,0x3B,0xD0,0xF2,0x4C,0xB9,0xF2,
0x85,0x31,0xBD,0x63,0x05,0x18,0x69,0x02,0x48,0xA5,0x31,0x20,0x38,0xF2,0xA0,0x00,0xB1,0x39,0x85,0x34,0xC8,0xB1,0x39,0x85,0x35,0xA9,0x00,0x09,0x1C,0xA6,0x33,0x85,
0x36,0xA9,0x02,0xA0,0x01,0x84,0x41,0x85,0x40,0xA9,0x00,0x85,0x42,0x68,0xAA,0xA5,0x34,0x85,0x3B,0xA5,0x40,0x4A,0x18,0x69,0x01,0xA8,0xB1,0x39,0x85,0x04,0xA5,0x40,
0x29,0x01,0xF0,0x07,0xA5,0x04,0x29,0x0F,0x4C,0x31,0xF3,0xA5,0x04,0x4A,0x4A,0x4A,0x4A,0x48,0xBD,0x95,0x05,0xA8,0x68,0x99,0x23,0x05,0xA5,0x36,0x99,0x63,0x04,0xE8,
0xA5,0x40,0x18,0x65,0x41,0x85,0x40,0xC6,0x3B,0xD0,0xC8,0xA5,0x40,0x18,0x65,0x42,0x85,0x40,0xC6,0x35,0xD0,0xB9,0x60,0xA5,0x43,0x49,0x01,0x85,0x43,0xD0,0x17,0xA0,
0x08,0xA2,0x3D,0xBD,0x63,0x04,0xF0,0x07,0x20,0x91,0xF3,0xCA,0x10,0xF5,0x60,0x20,0xAB,0xF3,0xCA,0x10,0xEE,0x60,0xA0,0x08,0xA2,0x00,0xBD,0x63,0x04,0xF0,0x09,0x20,
0x91,0xF3,0xE8,0xE0,0x3E,0xD0,0xF3,0x60,0x20,0xAB,0xF3,0xE8,0xE0,0x3E,0xD0,0xEA,0x60,0x99,0x02,0x02,0xBD,0xE3,0x04,0x99,0x00,0x02,0xBD,0x23,0x05,0x99,0x01,0x02,
0xBD,0xA3,0x04,0x99,0x03,0x02,0xC8,0xC8,0xC8,0xC8,0x60,0xA9,0xFF,0x99,0x00,0x02,0xC8,0xC8,0xC8,0xC8,0x60,0xA5,0x1F,0x29,0x01,0x0A,0x85,0x04,0xA5,0x1E,0x29,0x01,
0x05,0x04,0xA8,0xA5,0x1F,0x0A,0x0A,0x29,0xF8,0x85,0x04,0xA5,0x1E,0x4A,0x18,0x65,0x04,0xAA,0x09,0xC0,0x8D,0x42,0x04,0xA9,0x23,0x8D,0x41,0x04,0xA9,0x01,0x8D,0x40,
0x04,0xB9,0xFB,0xF3,0x25,0x1A,0x85,0x04,0xB9,0xFB,0xF3,0x49,0xFF,0x3D,0x00,0x04,0x05,0x04,0x9D,0x00,0x04,0x8D,0x43,0x04,0x4C,0xB9,0xF1,0x03,0x0C,0x30,0xC0,0x0D,
0x00,0x00,0x2C,0x00,0x00,0x00,0x26,0x00,0x00,0x00,0x20,0x00,0x00,0x00,0x00,0x0D,0x00,0x00,0x13,0x00,0x00,0x00,0x26,0x00,0x00,0x00,0x20,0x00,0x00,0x00,0x28,0xA9,
0x01,0x85,0x0C,0xA9,0x20,0x85,0x05,0x8D,0x06,0x20,0xA9,0x00,0x8D,0x06,0x20,0xA2,0x04,0xA8,0x8D,0x07,0x20,0x88,0xD0,0xFA,0xCA,0xD0,0xF7,0x20,0xD3,0xF5,0xA2,0x00,
0xA9,0x00,0x9D,0x63,0x05,0xE8,0xE0,0x32,0xD0,0xF8,0x20,0x69,0xF1,0xA9,0x0E,0x8D,0x50,0x00,0x85,0x2D,0xA9,0x3C,0x8D,0x51,0x00,0x85,0x2F,0xA9,0x80,0x8D,0x55,0x00,
0xA9,0x96,0x8D,0x56,0x00,0xA9,0x14,0x20,0xE9,0xF1,0x8D,0x4F,0x00,0xA9,0x02,0x20,0xE9,0xF1,0x8D,0x54,0x00,0xA9,0xFF,0x85,0x2F,0xA9,0x05,0x20,0xE9,0xF1,0x8D,0x7B,
0x00,0xA2,0x13,0xBD,0x43,0xF8,0x9D,0x63,0x04,0xCA,0x10,0xF7,0xA9,0x09,0x8D,0x15,0x40,0xA9,0x20,0x8D,0x40,0x04,0xA9,0x3F,0x8D,0x41,0x04,0xA9,0x00,0x8D,0x42,0x04,
0xA0,0x00,0xB9,0xFF,0xF3,0x99,0x43,0x04,0xC8,0xC0,0x20,0xD0,0xF5,0x20,0xB9,0xF1,0xA9,0x00,0x85,0x0C,0x20,0xC9,0xF4,0xA5,0x01,0x09,0x18,0x85,0x01,0x8D,0x01,0x20,
0x20,0xD4,0xF4,0x20,0xC9,0xF4,0x4C,0xC0,0xF4,0xA9,0x00,0x8D,0x49,0x00,0xAD,0x49,0x00,0xF0,0xFB,0x60,0x20,0x5C,0xF0,0x20,0x7B,0xF6,0x20,0x92,0xF9,0x20,0xEC,0xFA,
0x20,0x82,0xFC,0x20,0xD5,0xFC,0x20,0xDA,0xF8,0x20,0xF6,0xF8,0x20,0x89,0xFD,0xAD,0x50,0x00,0x85,0x2D,0xAD,0x51,0x00,0x85,0x2F,0xAE,0x4F,0x00,0x20,0x5B,0xF2,0x20,
0x57,0xF3,0x60,0xAD,0x5A,0x00,0xD0,0x5D,0xA5,0x07,0x29,0x08,0xF0,0x2A,0xA9,0x08,0x20,0xAC,0xF7,0xF0,0x23,0xA9,0x3F,0x8D,0x59,0x00,0xA9,0xF7,0x8D,0x5A,0x00,0xA9,
0x08,0x20,0x8F,0xF7,0xF0,0x05,0xA9,0x20,0x4C,0x2D,0xF5,0xA9,0x00,0x8D,0x5B,0x00,0xA9,0x08,0x8D,0x5E,0x00,0x4C,0x65,0xF5,0xA5,0x07,0x29,0x04,0xF0,0x27,0xA9,0x04,
0x20,0xAC,0xF7,0xF0,0x20,0xA9,0xFF,0x8D,0x59,0x00,0xA9,0xF6,0x8D,0x5A,0x00,0xA9,0x04,0x20,0x8F,0xF7,0xF0,0x05,0xA9,0x20,0x4C,0x5D,0xF5,0xA9,0x00,0x8D,0x5B,0x00,
0xA9,0x04,0x8D,0x5E,0x00,0xAD,0x58,0x00,0xD0,0x45,0xA5,0x07,0x29,0x02,0xF0,0x1E,0xA9,0x02,0x20,0xAC,0xF7,0xF0,0x17,0xA9,0x3F,0x8D,0x57,0x00,0xA9,0xF7,0x8D,0x58,
0x00,0xA9,0x00,0x8D,0x5C,0x00,0xA9,0x02,0x8D,0x5D,0x00,0x4C,0xAF,0xF5,0xA5,0x07,0x29,0x01,0xF0,0x1B,0xA9,0x01,0x20,0xAC,0xF7,0xF0,0x14,0xA9,0xFF,0x8D,0x57,0x00,
0xA9,0xF6,0x8D,0x58,0x00,0xA9,0x00,0x8D,0x5C,0x00,0xA9,0x01,0x8D,0x5D,0x00,0x60,0x85,0x04,0xA2,0xFF,0xE8,0xBD,0x64,0x05,0xDD,0x63,0x05,0xD0,0xF7,0x86,0x4B,0x85,
0x4A,0xE8,0xA5,0x04,0x18,0x7D,0x63,0x05,0x9D,0x63,0x05,0xE8,0xE0,0x32,0xD0,0xF2,0xA5,0x4B,0x60,0xA9,0x00,0x85,0x45,0xA9,0x04,0x85,0x46,0xA9,0x01,0xA2,0x05,0xA0,
0x03,0x20,0xA8,0xF0,0xA9,0x03,0x85,0x4C,0xA9,0x04,0x85,0x4D,0xA9,0x08,0x85,0x4E,0xA5,0x4C,0xA6,0x4D,0xA4,0x4E,0x20,0xA8,0xF0,0xA5,0x4D,0x18,0x69,0x04,0x85,0x4D,
0xC9,0x24,0xD0,0x0B,0xA9,0x04,0x85,0x4D,0xA5,0x4E,0x18,0x69,0x04,0x85,0x4E,0xE6,0x4C,0xA5,0x4C,0xC9,0x13,0xD0,0xD9,0xA9,0x03,0x85,0x4C,0xA9,0x04,0x85,0x4D,0xA9,
0x12,0x85,0x4E,0xA9,0x13,0xA6,0x4D,0xA4,0x4E,0x20,0xA8,0xF0,0xA5,0x4D,0x18,0x69,0x04,0x85,0x4D,0xC9,0x24,0xD0,0x0B,0xA9,0x04,0x85,0x4D,0xA5,0x4E,0x18,0x69,0x04,
0x85,0x4E,0xE6,0x4C,0xA5,0x4C,0xC9,0x1B,0xD0,0xD9,0xA9,0x55,0xA2,0x04,0xA0,0x00,0x20,0xB8,0xFC,0xA9,0x00,0xA2,0x04,0xA0,0x04,0x20,0xB8,0xFC,0xA9,0xAA,0xA2,0x02,
0xA0,0x09,0x20,0xB8,0xFC,0xA9,0xFF,0xA2,0x04,0xA0,0x0B,0x20,0xB8,0xFC,0xA2,0x00,0xA0,0x00,0x20,0xEA,0xF7,0xA9,0x02,0x8D,0x82,0x00,0x60,0xA5,0x5F,0x49,0x01,0x85,
0x5F,0xD0,0x03,0x20,0x03,0xF5,0xA5,0x58,0xF0,0x32,0xA4,0x5C,0xB1,0x57,0xC9,0x80,0xD0,0x07,0xA9,0x00,0x85,0x58,0x4C,0xA0,0xF6,0xA2,0x50,0x20,0x7F,0xF7,0xF0,0x1C,
0x98,0x29,0x0F,0xC9,0x07,0xD0,0x12,0x20,0x17,0xF8,0xA5,0x07,0x25,0x5D,0xF0,0x09,0xA5,0x5D,0x20,0xAC,0xF7,0xF0,0x02,0xA0,0x0F,0xC8,0x84,0x5C,0xA5,0x5A,0xF0,0x3E,
0xA4,0x5B,0xB1,0x59,0xC9,0x80,0xD0,0x07,0xA9,0x00,0x85,0x5A,0x4C,0xD6,0xF6,0xA2,0x51,0x20,0x7F,0xF7,0xF0,0x28,0x98,0x29,0x0F,0xC9,0x07,0xD0,0x1E,0x20,0x17,0xF8,
0xA5,0x07,0x25,0x5E,0xF0,0x15,0xA5,0x5E,0x20,0xAC,0xF7,0xF0,0x0E,0xA5,0x5E,0x20,0x8F,0xF7,0xF0,0x05,0xA0,0x2F,0x4C,0xFB,0xF6,0xA0,0x0F,0xC8,0x84,0x5B,0x60,0x01,
0x01,0x02,0x02,0x03,0x03,0x04,0x04,0x03,0x03,0x02,0x02,0x02,0x01,0xFF,0x80,0x05,0x04,0x04,0x03,0x03,0x04,0x05,0x04,0x03,0x03,0x02,0x02,0x02,0x01,0xFF,0x80,0x02,
0x03,0x04,0x05,0x06,0x06,0x06,0x04,0x03,0x03,0x02,0x02,0x02,0x01,0xFF,0x80,0x05,0x06,0x07,0x08,0x07,0x06,0x05,0x04,0x03,0x03,0x02,0x02,0x02,0x01,0xFF,0x80,0xFF,
0xFF,0xFE,0xFE,0xFD,0xFD,0xFC,0xFC,0xFD,0xFD,0xFE,0xFE,0xFE,0xFF,0x01,0x80,0xFB,0xFC,0xFC,0xFD,0xFD,0xFC,0xFB,0xFC,0xFD,0xFD,0xFE,0xFE,0xFE,0xFF,0x01,0x80,0xFE,
0xFD,0xFC,0xFB,0xFA,0xFA,0xFA,0xFC,0xFD,0xFD,0xFE,0xFE,0xFE,0xFF,0x01,0x80,0xFB,0xFA,0xF9,0xF8,0xF9,0xFA,0xFB,0xFC,0xFD,0xFD,0xFE,0xFE,0xFE,0xFF,0x01,0x80,0x48,
0x0A,0x68,0x6A,0xC6,0x5F,0xF0,0x01,0x18,0x75,0x00,0x95,0x00,0xE6,0x5F,0x60,0xC9,0x04,0xD0,0x09,0xA5,0x53,0xC9,0x02,0xD0,0x10,0xA9,0x01,0x60,0xC9,0x08,0xD0,0x09,
0xA5,0x53,0xC9,0x01,0xD0,0x03,0xA9,0x01,0x60,0xA9,0x00,0x60,0xC9,0x01,0xD0,0x0B,0xA5,0x52,0xC9,0x07,0xF0,0x31,0xE6,0x52,0x4C,0xE4,0xF7,0xC9,0x02,0xD0,0x09,0xA5,
0x52,0xF0,0x24,0xC6,0x52,0x4C,0xE4,0xF7,0xC9,0x04,0xD0,0x0B,0xA5,0x53,0xC9,0x04,0xF0,0x15,0xE6,0x53,0x4C,0xE4,0xF7,0xC9,0x08,0xD0,0x09,0xA5,0x53,0xF0,0x08,0xC6,
0x53,0x4C,0xE4,0xF7,0xA9,0x01,0x60,0xA9,0x00,0x60,0xA9,0xAA,0x85,0x1A,0x8A,0x0A,0x85,0x1E,0x98,0x0A,0x18,0x69,0x04,0xC9,0x08,0x90,0x03,0x18,0x69,0x01,0x85,0x1F,
0x20,0xB5,0xF3,0xE6,0x1E,0x20,0xB5,0xF3,0xE6,0x1F,0x20,0xB5,0xF3,0xC6,0x1E,0x4C,0xB5,0xF3,0xA9,0x00,0x4C,0xEC,0xF7,0x98,0x48,0xA6,0x60,0xA4,0x61,0x20,0x12,0xF8,
0xA5,0x53,0xC9,0x02,0xB0,0x14,0x85,0x61,0x85,0x63,0xA5,0x52,0x85,0x60,0x85,0x62,0xA6,0x60,0xA4,0x61,0x20,0xEA,0xF7,0x68,0xA8,0x60,0x85,0x63,0xA5,0x52,0x85,0x62,
0x68,0xA8,0x60,0x19,0x19,0x19,0x19,0x1A,0x19,0x19,0x19,0x19,0x1A,0x19,0x19,0x19,0x19,0x1A,0x19,0x19,0x19,0x19,0x1A,0xAD,0x77,0x04,0xC9,0x1A,0xD0,0x05,0xA9,0x18,
0x4C,0x65,0xF8,0xA9,0x1A,0xAE,0x6D,0x00,0xE0,0x07,0xB0,0x02,0x09,0x20,0x8D,0x77,0x04,0x8D,0x78,0x04,0x8D,0x79,0x04,0x8D,0x7A,0x04,0x60,0xA5,0x63,0x0A,0x0A,0x0A,
0x18,0x65,0x62,0x69,0x03,0x48,0xA5,0x65,0x0A,0x0A,0x18,0x69,0x12,0xA8,0xA5,0x64,0x0A,0x0A,0x18,0x69,0x04,0xAA,0x68,0x48,0x20,0xA8,0xF0,0x20,0x0A,0xFC,0x68,0x20,
0xD9,0xFB,0x4C,0xA5,0xF8,0xE6,0x64,0xA5,0x64,0xC9,0x06,0xD0,0x15,0xA5,0x65,0x0A,0x0A,0x0A,0x18,0x69,0x02,0xAA,0xBD,0x6B,0x06,0x38,0xE9,0x03,0x29,0x01,0xF0,0x06,
0xA5,0x64,0xC9,0x08,0xD0,0x11,0xA5,0x65,0xC9,0x02,0xF0,0x09,0xA9,0x00,0x85,0x64,0xE6,0x65,0x4C,0xD7,0xF8,0xC6,0x64,0x4C,0x0B,0xFD,0xA5,0x07,0x29,0x80,0xC5,0x66,
0xD0,0x01,0x60,0x85,0x66,0xC9,0x80,0xF0,0x01,0x60,0xA5,0x63,0xC9,0x02,0xB0,0x03,0x4C,0x7B,0xF8,0x4C,0x78,0xF9,0xA5,0x07,0x29,0x40,0xC5,0x67,0xD0,0x01,0x60,0x85,
0x67,0xC9,0x40,0xF0,0x01,0x60,0x4C,0x09,0xF9,0xA5,0x64,0xF0,0x25,0xA5,0x65,0xC9,0x02,0xD0,0x1A,0xAD,0x7D,0x06,0x38,0xE9,0x03,0x29,0x01,0xD0,0x04,0xA9,0x05,0xD0,
0x02,0xA9,0x07,0xC5,0x64,0xD0,0x06,0xAA,0xBD,0x7B,0x06,0xD0,0x0E,0x20,0xF9,0xFB,0xD0,0x26,0xA5,0x64,0x05,0x65,0xF0,0x20,0x20,0x5B,0xF9,0xA5,0x65,0x0A,0x0A,0x18,
0x69,0x12,0xA8,0xA5,0x64,0x0A,0x0A,0x18,0x69,0x04,0xAA,0x20,0xB5,0xFB,0xA9,0x13,0x20,0xA8,0xF0,0xA9,0x00,0x4C,0xD9,0xFB,0x4C,0x5B,0xF9,0xC6,0x64,0xA5,0x64,0xC9,
0xFF,0xD0,0x0F,0xA5,0x65,0xF0,0x09,0xA9,0x07,0x85,0x64,0xC6,0x65,0x4C,0x72,0xF9,0xE6,0x64,0x20,0xE2,0xFB,0x4C,0x0B,0xFD,0xA5,0x63,0x38,0xE9,0x02,0x85,0x65,0xA5,
0x62,0x85,0x64,0x20,0xE2,0xFB,0x8A,0x48,0x98,0x48,0x20,0x0B,0xFD,0x68,0xA8,0x68,0xAA,0x60,0xA5,0x65,0x0A,0x0A,0x0A,0x0A,0x0A,0x18,0x69,0x98,0x85,0x69,0xA5,0x64,
0x0A,0x0A,0x0A,0x0A,0x0A,0x18,0x69,0x0A,0x85,0x68,0xA5,0x55,0x8D,0x6E,0x00,0xA5,0x68,0x8D,0x6F,0x00,0x20,0x2F,0xFA,0x18,0x69,0x80,0x85,0x6A,0xA5,0x6B,0x20,0x56,
0xFA,0x85,0x6B,0xA5,0x56,0x8D,0x6E,0x00,0xA5,0x69,0x8D,0x6F,0x00,0x20,0x2F,0xFA,0x18,0x69,0x80,0x85,0x6A,0xA5,0x6C,0x20,0x56,0xFA,0x85,0x6C,0xA5,0x55,0x18,0x65,
0x6B,0x85,0x55,0xA5,0x56,0x18,0x65,0x6C,0x85,0x56,0xA6,0x6D,0xA5,0x55,0x18,0x7D,0x0D,0xFA,0x85,0x2D,0xA5,0x56,0x18,0x7D,0x1E,0xFA,0x85,0x2F,0xE8,0xE0,0x10,0xD0,
0x02,0xA2,0x00,0x86,0x6D,0xA6,0x54,0x20,0x5B,0xF2,0x4C,0x57,0xF8,0x00,0x04,0x07,0x09,0x0A,0x09,0x07,0x04,0x00,0xFC,0xF9,0xF7,0xF6,0xF7,0xF9,0xFC,0x00,0xF6,0xF7,
0xF9,0xFC,0x00,0x04,0x07,0x09,0x0A,0x09,0x07,0x04,0x00,0xFC,0xF9,0xF7,0xF6,0xA5,0x6E,0xC5,0x6F,0x90,0x12,0xD0,0x03,0xA9,0x00,0x60,0xA5,0x6F,0x38,0xE5,0x6E,0xA2,
0x03,0x38,0x6A,0xCA,0xD0,0xFB,0x60,0xA5,0x6F,0x38,0xE5,0x6E,0xA2,0x03,0x4A,0xCA,0xD0,0xFC,0x18,0x69,0x01,0x60,0x49,0x80,0xC5,0x6A,0x90,0x07,0xF0,0x07,0xE9,0x01,
0x4C,0x65,0xFA,0x69,0x01,0x49,0x80,0x60,0xA5,0x72,0xF0,0x0E,0xA9,0x00,0x85,0x72,0xA2,0x30,0x20,0xA8,0xFA,0xA9,0x01,0x85,0x74,0x60,0xA9,0x01,0x85,0x72,0xA2,0x20,
0x20,0xA8,0xFA,0xA9,0x01,0x85,0x73,0x60,0x00,0x06,0x08,0x06,0x00,0xFA,0xF8,0xFA,0x00,0x03,0x04,0x03,0x00,0xFD,0xFC,0xFD,0xF8,0xFA,0x00,0x06,0x08,0x06,0x00,0xFA,
0xFC,0xFD,0x00,0x03,0x04,0x03,0x00,0xFD,0xA0,0x00,0xB9,0x88,0xFA,0x18,0x65,0x70,0x9D,0xA3,0x04,0xB9,0x98,0xFA,0x18,0x65,0x71,0x9D,0xE3,0x04,0xB9,0x88,0xFA,0x9D,
0x0B,0x06,0xB9,0x98,0xFA,0x9D,0x2B,0x06,0xA9,0x01,0x9D,0x23,0x05,0xA9,0x1A,0x9D,0x63,0x04,0xE8,0xC8,0xC0,0x10,0xD0,0xD2,0xA9,0x0E,0x8D,0x0E,0x40,0xA9,0x04,0x8D,
0x0F,0x40,0xA9,0x25,0x8D,0x0C,0x40,0xA9,0x18,0x85,0x75,0x60,0xA5,0x75,0xF0,0x09,0xC6,0x75,0xD0,0x05,0xA9,0x30,0x8D,0x0C,0x40,0xA5,0x73,0xF0,0x0D,0xA2,0x20,0xE6,
0x73,0xA5,0x73,0xC9,0x18,0xF0,0x15,0x20,0x2D,0xFB,0xA5,0x74,0xF0,0x0D,0xA2,0x30,0xE6,0x74,0xA5,0x74,0xC9,0x18,0xF0,0x0E,0x20,0x2D,0xFB,0x60,0xA9,0x00,0x85,0x73,
0x20,0xA4,0xFB,0x4C,0x0A,0xFB,0xA9,0x00,0x85,0x74,0x4C,0xA4,0xFB,0x29,0x07,0x85,0x04,0xA0,0x10,0xBD,0xE3,0x04,0xC9,0xFF,0xF0,0x59,0xBD,0x63,0x04,0x49,0x02,0x9D,
0x63,0x04,0xBD,0x0B,0x06,0x10,0x09,0x18,0x7D,0xA3,0x04,0xB0,0x16,0x4C,0x56,0xFB,0x18,0x7D,0xA3,0x04,0x90,0x0D,0xA9,0xFF,0x9D,0xE3,0x04,0xA9,0x00,0x9D,0x63,0x04,
0x4C,0x93,0xFB,0x9D,0xA3,0x04,0xBD,0x2B,0x06,0x10,0x09,0x18,0x7D,0xE3,0x04,0xB0,0x09,0x4C,0x56,0xFB,0x18,0x7D,0xE3,0x04,0xB0,0xDC,0x9D,0xE3,0x04,0xA5,0x04,0xD0,
0x12,0xBD,0x0B,0x06,0x20,0x98,0xFB,0x9D,0x0B,0x06,0xBD,0x2B,0x06,0x20,0x98,0xFB,0x9D,0x2B,0x06,0xE8,0x88,0xD0,0x9C,0x60,0xF0,0x09,0x10,0x04,0x18,0x69,0x01,0x60,
0x38,0xE9,0x01,0x60,0xA0,0x10,0xA9,0x00,0x9D,0x63,0x04,0xA9,0xFF,0x9D,0xE3,0x04,0xE8,0x88,0xD0,0xF2,0x60,0x98,0x48,0x0A,0x0A,0x0A,0x18,0x69,0x0A,0x85,0x71,0x8A,
0x48,0x38,0xE9,0x04,0x0A,0x0A,0x0A,0x18,0x69,0x0D,0x85,0x70,0x20,0xF9,0xFB,0xF0,0x03,0x20,0x68,0xFA,0x68,0xAA,0x68,0xA8,0x60,0x48,0x20,0x00,0xFC,0x68,0x9D,0x6B,
0x06,0x60,0x20,0x00,0xFC,0xA5,0x64,0xF0,0x0F,0xBD,0x6B,0x06,0xD0,0x0A,0xCA,0xBD,0x6B,0x06,0xD0,0x04,0xC6,0x64,0xD0,0xF6,0x60,0x20,0x00,0xFC,0xBD,0x6B,0x06,0x60,
0xA5,0x65,0x0A,0x0A,0x0A,0x18,0x65,0x64,0xAA,0x60,0xA5,0x62,0x0A,0x0A,0x0A,0x0A,0x0A,0x85,0x76,0x85,0x2D,0xA5,0x63,0x0A,0x0A,0x0A,0x48,0x0A,0x0A,0x18,0x69,0x40,
0x85,0x77,0x85,0x2F,0x68,0x18,0x65,0x62,0x8D,0x7D,0x00,0x69,0x03,0xA6,0x7B,0x20,0xE0,0xF2,0xA6,0x7B,0x20,0x5B,0xF2,0xA5,0x65,0x0A,0x0A,0x0A,0x0A,0x0A,0x18,0x69,
0x90,0x38,0xE5,0x77,0x4A,0x4A,0x4A,0x4A,0x85,0x79,0xA5,0x64,0x38,0xE5,0x62,0x0A,0x85,0x78,0xA9,0x10,0x85,0x7A,0xAD,0x7D,0x00,0x0A,0x0A,0x0A,0x0A,0x85,0x04,0xA9,
0x00,0x06,0x04,0x2A,0x06,0x04,0x2A,0x18,0x69,0x02,0x8D,0x03,0x40,0xA5,0x04,0x8D,0x02,0x40,0xA9,0x24,0x8D,0x00,0x40,0xA9,0xF9,0x8D,0x01,0x40,0xA9,0x14,0x8D,0x7C,
0x00,0x60,0xA5,0x7C,0xF0,0x09,0xC6,0x7C,0xD0,0x05,0xA9,0x30,0x8D,0x00,0x40,0xA5,0x7A,0xD0,0x01,0x60,0xC6,0x7A,0xD0,0x09,0xA6,0x7B,0xA9,0xFF,0x85,0x2F,0x4C,0x5B,
0xF2,0xA5,0x76,0x18,0x65,0x78,0x85,0x76,0x85,0x2D,0xA5,0x77,0x18,0x65,0x79,0x85,0x77,0x85,0x2F,0xA6,0x7B,0x4C,0x5B,0xF2,0x85,0x1A,0x84,0x1F,0x86,0x7E,0xA9,0x00,
0x85,0x1E,0x20,0xB5,0xF3,0xA5,0x1E,0x18,0x69,0x01,0xC9,0x10,0xD0,0xF2,0xE6,0x1F,0xC6,0x7E,0xD0,0xEA,0x60,0xA6,0x7F,0xA4,0x80,0xC8,0xC0,0x05,0xD0,0x03,0xA0,0x00,
0xE8,0x84,0x80,0xE0,0x08,0xD0,0x02,0xA2,0x00,0x86,0x7F,0xBD,0x03,0xFD,0x8D,0x43,0x04,0xA9,0x01,0x8D,0x40,0x04,0xA9,0x3F,0x8D,0x41,0x04,0xA9,0x07,0x8D,0x42,0x04,
0x4C,0xB9,0xF1,0x21,0x2C,0x2B,0x28,0x27,0x25,0x24,0x2C,0xA6,0x81,0xA4,0x82,0x84,0x83,0xA9,0xAA,0x20,0xEC,0xF7,0xA5,0x65,0x18,0x69,0x02,0x85,0x82,0xA5,0x64,0x85,
0x81,0x20,0x3D,0xFD,0xA6,0x81,0xA4,0x82,0xA9,0x55,0x4C,0xEC,0xF7,0xAF,0xAF,0xAF,0xAF,0xAF,0xAF,0xAF,0xAF,0xFA,0xFA,0xFA,0xFA,0xFA,0xFA,0xFA,0xFA,0xA5,0x83,0xC5,
0x82,0xF0,0x45,0xA9,0x23,0x8D,0x41,0x04,0xA9,0xE0,0x8D,0x42,0x04,0xA0,0x1F,0xA9,0xFF,0x99,0x43,0x04,0x99,0x20,0x04,0x88,0x10,0xF5,0xA9,0x20,0x8D,0x40,0x04,0x20,
0xB9,0xF1,0xA5,0x82,0x38,0xE9,0x02,0x0A,0x0A,0x0A,0x18,0x69,0xE0,0x8D,0x42,0x04,0xA9,0x23,0x8D,0x41,0x04,0xA0,0x0F,0xB9,0x2D,0xFD,0x99,0x43,0x04,0x88,0x10,0xF7,
0xA9,0x10,0x8D,0x40,0x04,0x20,0xB9,0xF1,0x60,0xA5,0x07,0x29,0x30,0xD0,0x05,0xA9,0x01,0x85,0x84,0x60,0xA5,0x84,0xF0,0xFB,0xA9,0x00,0x8D,0x00,0x20,0x8D,0x01,0x20,
0xA9,0x6B,0x85,0x80,0xA9,0x06,0x85,0x81,0xA9,0x90,0x85,0x82,0xA9,0x00,0x85,0x83,0xA2,0x0F,0xA9,0xFF,0x95,0x90,0xCA,0x10,0xFB,0xA9,0x03,0x85,0x88,0xA9,0xEF,0x85,
0x86,0xA9,0x02,0x85,0x87,0xA9,0x71,0x85,0x89,0xA0,0x00,0x84,0x84,0xB1,0x80,0xF0,0x17,0x38,0xE9,0x03,0x29,0x0F,0x4A,0x05,0x84,0x91,0x80,0xA9,0x00,0x2A,0x0A,0x0A,
0x0A,0x85,0x84,0xC8,0xC0,0x08,0xD0,0xE5,0x84,0x85,0xC0,0x08,0xF0,0x07,0xC0,0x06,0xF0,0x0A,0x4C,0x70,0xFE,0xA0,0x03,0xB1,0x80,0x4C,0x00,0xFE,0xA0,0x03,0xB1,0x80,
0xA0,0x00,0xB1,0x80,0x29,0x07,0x05,0x84,0x91,0x80,0xA2,0x00,0x86,0x84,0xBC,0xB6,0xFE,0xB1,0x80,0x0A,0x0A,0x0A,0x0A,0xE8,0xBC,0xB6,0xFE,0x11,0x80,0xA4,0x84,0x99,
0x8A,0x00,0xE6,0x84,0xE8,0xE0,0x08,0xD0,0xE5,0xA5,0x8A,0x29,0x7F,0x85,0x8A,0xA5,0x8A,0xA6,0x8B,0xC5,0x90,0xD0,0x04,0xE4,0x91,0xF0,0x35,0xC5,0x94,0xD0,0x04,0xE4,
0x95,0xF0,0x2D,0xC5,0x98,0xD0,0x04,0xE4,0x99,0xF0,0x25,0xA0,0x01,0xB9,0x8A,0x00,0x91,0x82,0x88,0x10,0xF8,0xA0,0x03,0xB9,0x89,0x00,0x91,0x82,0x88,0xB9,0x8B,0x00,
0x91,0x82,0xA5,0x89,0x25,0x86,0xA6,0x85,0xE0,0x08,0xD0,0x02,0x05,0x87,0x85,0x89,0x38,0x26,0x86,0x06,0x87,0xA5,0x80,0x18,0x69,0x08,0x85,0x80,0x90,0x02,0xE6,0x81,
0xA5,0x82,0x18,0x69,0x04,0x85,0x82,0x90,0x02,0xE6,0x83,0xC6,0x88,0xF0,0x03,0x4C,0xC9,0xFD,0xA2,0x18,0xBD,0x9F,0xFE,0x95,0x10,0xCA,0x10,0xF8,0x4C,0x10,0x00,0xA2,
0x0B,0xB5,0x90,0x9D,0x01,0x80,0xCA,0x10,0xF8,0xA5,0x89,0x8D,0x00,0x80,0xA9,0x00,0x8D,0x00,0x80,0x6C,0xFC,0xFF,0x03,0x05,0x02,0x04,0x01,0x00,0x07,0x06,0x00,0x2A,
0x00,0x30,0x00,0x6E,0x00,0x72,0x00,0x7C,0x00,0x86,0x00,0x90,0x00,0x9A,0x00,0xA4,0x00,0xAE,0x00,0xB8,0x00,0xC2,0x00,0xCC,0x00,0xD6,0x00,0xE0,0x00,0xEA,0x00,0xF4,
0x00,0xFE,0x01,0x08,0x01,0x12,0x01,0x1C,0x08,0x02,0x01,0x23,0x45,0x67,0x1E,0x04,0x2C,0xC0,0x00,0x00,0x00,0x00,0x00,0x02,0xCC,0x00,0x00,0x00,0x08,0x00,0x00,0x52,
0x32,0xCC,0x5D,0x96,0x2C,0xE4,0x05,0x23,0x2C,0xE4,0xDC,0x6A,0x2C,0xE4,0x60,0x2A,0x00,0x55,0x5A,0xA0,0x40,0x06,0x02,0xA0,0x40,0x50,0xAA,0xA0,0x40,0x0C,0x40,0xC0,
0x44,0x48,0x0C,0xC4,0x00,0xC4,0x0C,0xC4,0x40,0x88,0x0C,0xC4,0x02,0x02,0xB1,0x80,0x04,0x04,0x0A,0x00,0x05,0x50,0xAC,0xE0,0xC4,0xC4,0x04,0x04,0x8D,0xE0,0x07,0x10,
0x05,0x20,0x8C,0xC0,0x04,0x04,0x8D,0x60,0x07,0x90,0x05,0x00,0x8C,0x00,0x04,0x04,0x09,0x60,0xA0,0x05,0x81,0x24,0x08,0x40,0x04,0x04,0xAC,0xE0,0x02,0x40,0x24,0x20,
0x8C,0xC0,0x04,0x04,0xE4,0xE4,0x06,0x40,0x24,0x60,0xC4,0xC4,0x04,0x04,0x8D,0x00,0x05,0x00,0x05,0x20,0x8C,0xC0,0x04,0x04,0xE4,0xE4,0xA0,0xA0,0xA0,0xA0,0x0C,0x40,
0x04,0x04,0x09,0xC5,0xA0,0x31,0x81,0x05,0x08,0xC0,0x04,0x04,0x8D,0xA4,0x07,0x40,0x05,0x60,0x8C,0x84,0x04,0x04,0x0E,0x40,0x0A,0x00,0x0A,0x00,0x0C,0x40,0x04,0x04,
0x2C,0x70,0x87,0x10,0x20,0xE0,0x0C,0x40,0x04,0x04,0xAE,0xE0,0x0A,0x00,0x0A,0x00,0x0C,0x40,0x04,0x04,0xE4,0xE4,0x81,0x90,0x06,0x40,0x08,0x00,0x04,0x04,0xE4,0xE4,
0x06,0x40,0x0A,0x00,0x0C,0x40,0x04,0x04,0xE0,0xE4,0xA6,0xA0,0xA0,0xE0,0xC4,0x80,0x04,0x04,0x00,0x00,0x23,0x31,0x00,0x00,0x00,0x00,0x05,0x04,0x00,0x70,0x03,0x3B,
0x71,0x0E,0xFF,0x50,0x0C,0xC4,0x03,0x3B,0x71,0x0E,0xFF,0x50,0x0C,0xC4,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xAA,0x7D,0xF0,0x00,0xF0,0xFF,0xFF};
unsigned char GameGenieCHR[0x400] = {
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xF0,0xF0,0xF0,0xF0,0x00,0x00,0x00,0x00,0xF0,0xF0,0xF0,0xF0,
0x00,0x00,0x00,0x00,0x0F,0x0F,0x0F,0x0F,0x00,0x00,0x00,0x00,0x0F,0x0F,0x0F,0x0F,0x00,0x00,0x00,0x00,0xFF,0xFF,0xFF,0xFF,0x00,0x00,0x00,0x00,0xFF,0xFF,0xFF,0xFF,
0xF0,0xF0,0xF0,0xF0,0x00,0x00,0x00,0x00,0xF0,0xF0,0xF0,0xF0,0x00,0x00,0x00,0x00,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,
0xF0,0xF0,0xF0,0xF0,0x0F,0x0F,0x0F,0x0F,0xF0,0xF0,0xF0,0xF0,0x0F,0x0F,0x0F,0x0F,0xF0,0xF0,0xF0,0xF0,0xFF,0xFF,0xFF,0xFF,0xF0,0xF0,0xF0,0xF0,0xFF,0xFF,0xFF,0xFF,
0x0F,0x0F,0x0F,0x0F,0x00,0x00,0x00,0x00,0x0F,0x0F,0x0F,0x0F,0x00,0x00,0x00,0x00,0x0F,0x0F,0x0F,0x0F,0xF0,0xF0,0xF0,0xF0,0x0F,0x0F,0x0F,0x0F,0xF0,0xF0,0xF0,0xF0,
0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0xFF,0xFF,0xFF,0xFF,0x0F,0x0F,0x0F,0x0F,0xFF,0xFF,0xFF,0xFF,
0xFF,0xFF,0xFF,0xFF,0x00,0x00,0x00,0x00,0xFF,0xFF,0xFF,0xFF,0x00,0x00,0x00,0x00,0xFF,0xFF,0xFF,0xFF,0xF0,0xF0,0xF0,0xF0,0xFF,0xFF,0xFF,0xFF,0xF0,0xF0,0xF0,0xF0,
0xFF,0xFF,0xFF,0xFF,0x0F,0x0F,0x0F,0x0F,0xFF,0xFF,0xFF,0xFF,0x0F,0x0F,0x0F,0x0F,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xF0,0xF0,0xF0,0xF0,0x00,0x00,0x00,0x00,0xF0,0xF0,0xF0,0xF0,
0x00,0x00,0x00,0x00,0x0F,0x0F,0x0F,0x0F,0x00,0x00,0x00,0x00,0x0F,0x0F,0x0F,0x0F,0x00,0x00,0x00,0x00,0xFF,0xFF,0xFF,0xFF,0x00,0x00,0x00,0x00,0xFF,0xFF,0xFF,0xFF,
0xF0,0xF0,0xF0,0xF0,0x00,0x00,0x00,0x00,0xF0,0xF0,0xF0,0xF0,0x00,0x00,0x00,0x00,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,
0xF0,0xF0,0xF0,0xF0,0x0F,0x0F,0x0F,0x0F,0xF0,0xF0,0xF0,0xF0,0x0F,0x0F,0x0F,0x0F,0xF0,0xF0,0xF0,0xF0,0xFF,0xFF,0xFF,0xFF,0xF0,0xF0,0xF0,0xF0,0xFF,0xFF,0xFF,0xFF,
0x0F,0x0F,0x0F,0x0F,0x00,0x00,0x00,0x00,0x0F,0x0F,0x0F,0x0F,0x00,0x00,0x00,0x00,0x0F,0x0F,0x0F,0x0F,0xF0,0xF0,0xF0,0xF0,0x0F,0x0F,0x0F,0x0F,0xF0,0xF0,0xF0,0xF0,
0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0xFF,0xFF,0xFF,0xFF,0x0F,0x0F,0x0F,0x0F,0xFF,0xFF,0xFF,0xFF,
0xFF,0xFF,0xFF,0xFF,0x00,0x00,0x00,0x00,0xFF,0xFF,0xFF,0xFF,0x00,0x00,0x00,0x00,0xFF,0xFF,0xFF,0xFF,0xF0,0xF0,0xF0,0xF0,0xFF,0xFF,0xFF,0xFF,0xF0,0xF0,0xF0,0xF0,
0xFF,0xFF,0xFF,0xFF,0x0F,0x0F,0x0F,0x0F,0xFF,0xFF,0xFF,0xFF,0x0F,0x0F,0x0F,0x0F,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xF0,0xF0,0xF0,0xF0,0x00,0x00,0x00,0x00,0xF0,0xF0,0xF0,0xF0,
0x00,0x00,0x00,0x00,0x0F,0x0F,0x0F,0x0F,0x00,0x00,0x00,0x00,0x0F,0x0F,0x0F,0x0F,0x00,0x00,0x00,0x00,0xFF,0xFF,0xFF,0xFF,0x00,0x00,0x00,0x00,0xFF,0xFF,0xFF,0xFF,
0xF0,0xF0,0xF0,0xF0,0x00,0x00,0x00,0x00,0xF0,0xF0,0xF0,0xF0,0x00,0x00,0x00,0x00,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,
0xF0,0xF0,0xF0,0xF0,0x0F,0x0F,0x0F,0x0F,0xF0,0xF0,0xF0,0xF0,0x0F,0x0F,0x0F,0x0F,0xF0,0xF0,0xF0,0xF0,0xFF,0xFF,0xFF,0xFF,0xF0,0xF0,0xF0,0xF0,0xFF,0xFF,0xFF,0xFF,
0x0F,0x0F,0x0F,0x0F,0x00,0x00,0x00,0x00,0x0F,0x0F,0x0F,0x0F,0x00,0x00,0x00,0x00,0x0F,0x0F,0x0F,0x0F,0xF0,0xF0,0xF0,0xF0,0x0F,0x0F,0x0F,0x0F,0xF0,0xF0,0xF0,0xF0,
0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0xFF,0xFF,0xFF,0xFF,0x0F,0x0F,0x0F,0x0F,0xFF,0xFF,0xFF,0xFF,
0xFF,0xFF,0xFF,0xFF,0x00,0x00,0x00,0x00,0xFF,0xFF,0xFF,0xFF,0x00,0x00,0x00,0x00,0xFF,0xFF,0xFF,0xFF,0xF0,0xF0,0xF0,0xF0,0xFF,0xFF,0xFF,0xFF,0xF0,0xF0,0xF0,0xF0,
0xFF,0xFF,0xFF,0xFF,0x0F,0x0F,0x0F,0x0F,0xFF,0xFF,0xFF,0xFF,0x0F,0x0F,0x0F,0x0F,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xF0,0xF0,0xF0,0xF0,0x00,0x00,0x00,0x00,0xF0,0xF0,0xF0,0xF0,
0x00,0x00,0x00,0x00,0x0F,0x0F,0x0F,0x0F,0x00,0x00,0x00,0x00,0x0F,0x0F,0x0F,0x0F,0x00,0x00,0x00,0x00,0xFF,0xFF,0xFF,0xFF,0x00,0x00,0x00,0x00,0xFF,0xFF,0xFF,0xFF,
0xF0,0xF0,0xF0,0xF0,0x00,0x00,0x00,0x00,0xF0,0xF0,0xF0,0xF0,0x00,0x00,0x00,0x00,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,
0xF0,0xF0,0xF0,0xF0,0x0F,0x0F,0x0F,0x0F,0xF0,0xF0,0xF0,0xF0,0x0F,0x0F,0x0F,0x0F,0xF0,0xF0,0xF0,0xF0,0xFF,0xFF,0xFF,0xFF,0xF0,0xF0,0xF0,0xF0,0xFF,0xFF,0xFF,0xFF,
0x0F,0x0F,0x0F,0x0F,0x00,0x00,0x00,0x00,0x0F,0x0F,0x0F,0x0F,0x00,0x00,0x00,0x00,0x0F,0x0F,0x0F,0x0F,0xF0,0xF0,0xF0,0xF0,0x0F,0x0F,0x0F,0x0F,0xF0,0xF0,0xF0,0xF0,
0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0xFF,0xFF,0xFF,0xFF,0x0F,0x0F,0x0F,0x0F,0xFF,0xFF,0xFF,0xFF,
0xFF,0xFF,0xFF,0xFF,0x00,0x00,0x00,0x00,0xFF,0xFF,0xFF,0xFF,0x00,0x00,0x00,0x00,0xFF,0xFF,0xFF,0xFF,0xF0,0xF0,0xF0,0xF0,0xFF,0xFF,0xFF,0xFF,0xF0,0xF0,0xF0,0xF0,
0xFF,0xFF,0xFF,0xFF,0x0F,0x0F,0x0F,0x0F,0xFF,0xFF,0xFF,0xFF,0x0F,0x0F,0x0F,0x0F,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF};

static int Code1B, Code2B, Code3B;
static int Code1A, Code2A, Code3A;
static int Code1O, Code2O, Code3O;
static int Code1V, Code2V, Code3V;
static unsigned char CodeStat;

int	__cdecl	GenieReader (int Bank, int Addy)
{
	int result = CPU_ReadPRG(Bank,Addy);
	if (NES.GameGenie)
	{
		if ((CodeStat & 0x10) && (Bank == Code1B) && (Addy == Code1A) && ((CodeStat & 0x02) || (result == Code1O)))
			return Code1V;
		if ((CodeStat & 0x20) && (Bank == Code2B) && (Addy == Code2A) && ((CodeStat & 0x04) || (result == Code2O)))
			return Code2V;
		if ((CodeStat & 0x40) && (Bank == Code3B) && (Addy == Code3A) && ((CodeStat & 0x08) || (result == Code3O)))
			return Code3V;
	}
	return result;
}

void	__cdecl	GenieWriter (int Bank, int Addy, int Val)
{
	int i;
	switch (Addy)
	{
	case 0:	if (Val & 1)
			CodeStat = Val ^ 0x7E;
		else
		{
			MP.SetCHR_RAM8(0,0);
			MP.Mirror_4();
			for (i = 0x8; i < 0x10; i++)
			{
				CPU.WriteHandler[i] = CPU_WritePRG;
				CPU.ReadHandler[i] = GenieReader;
				CPU.Readable[i] = FALSE;
			}
			MI = MI2;
			PPU_GetHandlers();
			if (MI)
				MI->InitMapper(&MP,0);
		}
		break;
	case 1:	Code1B = 8 | (Val >> 4);
		Code1A = (Code1A & 0xFF) | ((Val << 8) & 0xF00);
		break;
	case 2:	Code1A = (Code1A & 0xF00) | (Val);
		break;
	case 3:	Code1O = Val;
		break;
	case 4:	Code1V = Val;
		break;
	case 5:	Code2B = 8 | (Val >> 4);
		Code2A = (Code2A & 0xFF) | ((Val << 8) & 0xF00);
		break;
	case 6:	Code2A = (Code2A & 0xF00) | (Val);
		break;
	case 7:	Code2O = Val;
		break;
	case 8:	Code2V = Val;
		break;
	case 9:	Code3B = 8 | (Val >> 4);
		Code3A = (Code3A & 0xFF) | ((Val << 8) & 0xF00);
		break;
	case 10:Code3A = (Code3A & 0xF00) | (Val);
		break;
	case 11:Code3O = Val;
		break;
	case 12:Code3V = Val;
		break;
	}
}

void	NES_Reset (int ResetType)
{
	int i;
	CPU.ReadHandler[0] = CPU_ReadRAM;	CPU.WriteHandler[0] = CPU_WriteRAM;
	CPU.ReadHandler[1] = CPU_ReadRAM;	CPU.WriteHandler[1] = CPU_WriteRAM;
	CPU.ReadHandler[2] = PPU_Read;		CPU.WriteHandler[2] = PPU_Write;
	CPU.ReadHandler[3] = PPU_Read;		CPU.WriteHandler[3] = PPU_Write;
	CPU.ReadHandler[4] = CPU_Read4k;	CPU.WriteHandler[4] = CPU_Write4k;
	for (i = 0x5; i < 0x10; i++)
	{
		CPU.ReadHandler[i] = CPU_ReadPRG;
		CPU.WriteHandler[i] = CPU_WritePRG;
		CPU.Readable[i] = FALSE;
		CPU.Writable[i] = FALSE;
	}
	switch (ResetType)
	{
	case 0:	if (!NES.IsNSF)
			ZeroMemory(PRG_RAM,sizeof(PRG_RAM));
		ZeroMemory(CHR_RAM,sizeof(CHR_RAM));
		PPU_PowerOn();
		CPU_PowerOn();
		MP.SetCHR_RAM8(0,0);
		MP.Mirror_4();
		if (NES.IsNSF)
		{
			MP.SetPRG_RAM4(0x5,2);
			MP.SetPRG_RAM8(0x6,0);
			CPU.ReadHandler[0xF] = NSFReader;
		}
		if (MI)
			MI->InitMapper(&MP,1);
		break;
	case RESET_HARD:
		if ((MI) && (MI->UnloadMapper))
			MI->UnloadMapper();
		if (!NES.IsNSF)
			ZeroMemory(PRG_RAM,sizeof(PRG_RAM));
		ZeroMemory(CHR_RAM,sizeof(CHR_RAM));
		CPU_PowerOn();
		if (MI2)
		{
			MI = MI2;
			MI2 = NULL;
		}
		PPU_PowerOn();
		if ((NES.GameGenie) && (!NES.IsNSF))
		{
			Code1B = Code2B = Code3B = 0;
			Code1A = Code2A = Code3A = 0;
			Code1O = Code2O = Code3O = 0;
			Code1V = Code2V = Code3V = 0;
			CodeStat = 0;
			for (i = 0x8; i < 0x10; i++)
			{
				CPU.Readable[i] = TRUE;
				CPU.PRGPointer[i] = GameGeniePRG;
			}
			CPU.WriteHandler[0x8] = GenieWriter;
			for (i = 0; i < 8; i++)
			{
				PPU.Writable[i] = FALSE;
				PPU.CHRPointer[i] = GameGenieCHR;
			}
			MP.Mirror_V();
			MI2 = MI;
			MI = NULL;
			PPU_GetHandlers();
			break;
		}
		MP.SetCHR_RAM8(0,0);
		MP.Mirror_4();
		if (NES.IsNSF)
		{
			MP.SetPRG_RAM4(0x5,2);
			MP.SetPRG_RAM8(0x6,0);
			CPU.ReadHandler[0xF] = NSFReader;
		}
		if (MI)
			MI->InitMapper(&MP,1);
		break;
	case RESET_SOFT:
		if ((MI) && (MI->UnloadMapper))
			MI->UnloadMapper();
		MP.SetCHR_RAM8(0,0);
		MP.Mirror_4();
		if (NES.IsNSF)
		{
			MP.SetPRG_RAM4(0x5,2);
			MP.SetPRG_RAM8(0x6,0);
			CPU.ReadHandler[0xF] = NSFReader;
		}
		if (MI)
			MI->InitMapper(&MP,0);
		break;
	}
	Debugger.TraceOffset = -1;
	PPU.Clockticks = -1;
	PPU.SLnum = 0;
	APU_Reset();
	CPU_Reset();
	if (Debugger.Enabled)
		Debugger_Update();
}

void	NES_Run (void)
{
	NES.Stopped = FALSE;

	Debugger.TraceOffset = -1;

	Controllers_Acquire();
	if ((!NES.Stop) && (NES.SoundEnabled))
		APU_SoundON();

	NES.NeedReset = 0;
	do
	{
		CPU_ExecOp();
		if (Debugger.Enabled)
		{
			Debugger_AddInst();
			Debugger_Update();
		}
		if (NES.Scanline)
		{
			int SLnum = PPU.SLnum;
			NES.Scanline = FALSE;
			ProcessMessages();
			if (SLnum == 0)
			{
				if (States.NeedSave)
					States_SaveState();
				if (States.NeedLoad)
					States_LoadState();
			}
			else if ((SLnum == 240) && (Debugger.Enabled))
				Debugger_UpdateGraphics();
			else if (SLnum == 241)
				Controllers_UpdateInput();
		}
	}	while (!NES.Stop);

	APU_SoundOFF();
	Controllers_UnAcquire();
	GFX_UpdateTitlebar();

	NES.Stopped = TRUE;

	if (!NES.ROMLoaded)
	{
		NES_CloseFile();
		return;
	}
	if (NES.NeedQuit)
	{
		NES_Release();
		return;
	}
	if (NES.NeedReset == RESET_HARD)
	{
		NES.NeedReset = 0;
		NES_Reset(RESET_HARD);
	}
	else if (NES.NeedReset == RESET_SOFT)
	{
		NES.NeedReset = 0;
		NES_Reset(RESET_SOFT);
		NES.Stop = FALSE;
		NES_Run();
	}
}

void	NES_UpdateInterface (void)
{
	int W, H;
	Debugger.DoubleSize = (SizeMult > 1) ? 1 : 0;
	Debugger_SetMode(Debugger.Mode);
	W = 256 * SizeMult;
	H = 240 * SizeMult;
	SetWindowPos(mWnd,HWND_TOP,0,0,W,H,SWP_NOMOVE);
	W += (W - Width);
	H += (H - Height);
	SetWindowPos(mWnd,HWND_TOP,0,0,W,H,SWP_NOMOVE);
}

void	NES_LoadSettings (void)
{
	HMENU hMenu = GetMenu(mWnd);
	HKEY SettingsBase;
	unsigned long Size = 4;	unsigned long Type = REG_DWORD;
	int Port1T = 0, Port2T = 0, FSPort1T = 0, FSPort2T = 0, FSPort3T = 0, FSPort4T = 0, ExpPortT = 0;

	//Load Defaults
	SizeMult = 1;
	NES.SoundEnabled = 1;
	GFX.aFSkip = 1;
	GFX.FSkip = 0;
	Controllers.Port1.Type = 0;
	ZeroMemory(Controllers.Port1.Buttons,sizeof(Controllers.Port1.Buttons));
	Controllers.Port2.Type = 0;
	ZeroMemory(Controllers.Port2.Buttons,sizeof(Controllers.Port2.Buttons));
	Controllers.ExpPort.Type = 0;
	ZeroMemory(Controllers.ExpPort.Buttons,sizeof(Controllers.ExpPort.Buttons));
	//End Defaults

	RegOpenKeyEx(HKEY_CURRENT_USER, "SOFTWARE\\Nintendulator\\", 0, KEY_ALL_ACCESS, &SettingsBase);
	RegQueryValueEx(SettingsBase,"SoundEnabled",0,&Type,(unsigned char *)&NES.SoundEnabled,&Size);
	RegQueryValueEx(SettingsBase,"SizeMult"    ,0,&Type,(unsigned char *)&SizeMult  ,&Size);
	RegQueryValueEx(SettingsBase,"FSkip"       ,0,&Type,(unsigned char *)&GFX.FSkip ,&Size);
	RegQueryValueEx(SettingsBase,"aFSkip"      ,0,&Type,(unsigned char *)&GFX.aFSkip,&Size);
	RegQueryValueEx(SettingsBase,"PPUMode"     ,0,&Type,(unsigned char *)&PPU.IsPAL ,&Size);

	RegQueryValueEx(SettingsBase,"Port1T"  ,0,&Type,(unsigned char *)&Port1T  ,&Size);
	RegQueryValueEx(SettingsBase,"Port2T"  ,0,&Type,(unsigned char *)&Port2T  ,&Size);
	RegQueryValueEx(SettingsBase,"FSPort1T",0,&Type,(unsigned char *)&FSPort1T,&Size);
	RegQueryValueEx(SettingsBase,"FSPort2T",0,&Type,(unsigned char *)&FSPort2T,&Size);
	RegQueryValueEx(SettingsBase,"FSPort3T",0,&Type,(unsigned char *)&FSPort3T,&Size);
	RegQueryValueEx(SettingsBase,"FSPort4T",0,&Type,(unsigned char *)&FSPort4T,&Size);
	RegQueryValueEx(SettingsBase,"ExpPortT",0,&Type,(unsigned char *)&ExpPortT,&Size);


	if ((Port1T == 5) || (Port2T == 5))
		StdPort_SetControllerType(&Controllers.Port1,5);
	else
	{
		StdPort_SetControllerType(&Controllers.Port1,Port1T);
		StdPort_SetControllerType(&Controllers.Port2,Port2T);
	}
	StdPort_SetControllerType(&Controllers.FSPort1,FSPort1T);
	StdPort_SetControllerType(&Controllers.FSPort2,FSPort2T);
	StdPort_SetControllerType(&Controllers.FSPort3,FSPort3T);
	StdPort_SetControllerType(&Controllers.FSPort4,FSPort4T);
	ExpPort_SetControllerType(&Controllers.ExpPort,ExpPortT);

	Type = REG_BINARY;
	Size = sizeof(Controllers.Port1.Buttons);
	RegQueryValueEx(SettingsBase,"Port1D"  ,0,&Type,(unsigned char *)Controllers.Port1.Buttons  ,&Size);
	RegQueryValueEx(SettingsBase,"Port2D"  ,0,&Type,(unsigned char *)Controllers.Port2.Buttons  ,&Size);
	RegQueryValueEx(SettingsBase,"FSPort1D",0,&Type,(unsigned char *)Controllers.FSPort1.Buttons,&Size);
	RegQueryValueEx(SettingsBase,"FSPort2D",0,&Type,(unsigned char *)Controllers.FSPort2.Buttons,&Size);
	RegQueryValueEx(SettingsBase,"FSPort3D",0,&Type,(unsigned char *)Controllers.FSPort3.Buttons,&Size);
	RegQueryValueEx(SettingsBase,"FSPort4D",0,&Type,(unsigned char *)Controllers.FSPort4.Buttons,&Size);
	Size = sizeof(Controllers.ExpPort.Buttons);
	RegQueryValueEx(SettingsBase,"ExpPortD",0,&Type,(unsigned char *)Controllers.ExpPort.Buttons,&Size);
	Controllers_SetDeviceUsed();
	
	RegCloseKey(SettingsBase);
	
	if (NES.SoundEnabled)
		CheckMenuItem(hMenu, ID_SOUND_ENABLED, MF_CHECKED);
	
	if (GFX.aFSkip)
		CheckMenuItem(hMenu, ID_PPU_FRAMESKIP_AUTO, MF_CHECKED);
	switch (GFX.FSkip)
	{
	case 0:	CheckMenuRadioItem(hMenu,ID_PPU_FRAMESKIP_0,ID_PPU_FRAMESKIP_9,ID_PPU_FRAMESKIP_0,MF_BYCOMMAND);	break;
	case 1:	CheckMenuRadioItem(hMenu,ID_PPU_FRAMESKIP_0,ID_PPU_FRAMESKIP_9,ID_PPU_FRAMESKIP_1,MF_BYCOMMAND);	break;
	case 2:	CheckMenuRadioItem(hMenu,ID_PPU_FRAMESKIP_0,ID_PPU_FRAMESKIP_9,ID_PPU_FRAMESKIP_2,MF_BYCOMMAND);	break;
	case 3:	CheckMenuRadioItem(hMenu,ID_PPU_FRAMESKIP_0,ID_PPU_FRAMESKIP_9,ID_PPU_FRAMESKIP_3,MF_BYCOMMAND);	break;
	case 4:	CheckMenuRadioItem(hMenu,ID_PPU_FRAMESKIP_0,ID_PPU_FRAMESKIP_9,ID_PPU_FRAMESKIP_4,MF_BYCOMMAND);	break;
	case 5:	CheckMenuRadioItem(hMenu,ID_PPU_FRAMESKIP_0,ID_PPU_FRAMESKIP_9,ID_PPU_FRAMESKIP_5,MF_BYCOMMAND);	break;
	case 6:	CheckMenuRadioItem(hMenu,ID_PPU_FRAMESKIP_0,ID_PPU_FRAMESKIP_9,ID_PPU_FRAMESKIP_6,MF_BYCOMMAND);	break;
	case 7:	CheckMenuRadioItem(hMenu,ID_PPU_FRAMESKIP_0,ID_PPU_FRAMESKIP_9,ID_PPU_FRAMESKIP_7,MF_BYCOMMAND);	break;
	case 8:	CheckMenuRadioItem(hMenu,ID_PPU_FRAMESKIP_0,ID_PPU_FRAMESKIP_9,ID_PPU_FRAMESKIP_8,MF_BYCOMMAND);	break;
	case 9:	CheckMenuRadioItem(hMenu,ID_PPU_FRAMESKIP_0,ID_PPU_FRAMESKIP_9,ID_PPU_FRAMESKIP_9,MF_BYCOMMAND);	break;
	}
	Debugger.DoubleSize = (SizeMult > 1) ? 1 : 0;

	switch (SizeMult)
	{
	case 1:	CheckMenuRadioItem(hMenu,ID_PPU_SIZE_1X,ID_PPU_SIZE_4X,ID_PPU_SIZE_1X,MF_BYCOMMAND);	break;
	case 2:	CheckMenuRadioItem(hMenu,ID_PPU_SIZE_1X,ID_PPU_SIZE_4X,ID_PPU_SIZE_2X,MF_BYCOMMAND);	break;
	case 3:	CheckMenuRadioItem(hMenu,ID_PPU_SIZE_1X,ID_PPU_SIZE_4X,ID_PPU_SIZE_3X,MF_BYCOMMAND);	break;
	case 4:	CheckMenuRadioItem(hMenu,ID_PPU_SIZE_1X,ID_PPU_SIZE_4X,ID_PPU_SIZE_4X,MF_BYCOMMAND);	break;
	default:SizeMult = 1;
		CheckMenuRadioItem(hMenu,ID_PPU_SIZE_1X,ID_PPU_SIZE_4X,ID_PPU_SIZE_1X,MF_BYCOMMAND);	break;
	}

	if (PPU.IsPAL)
		NES_SetCPUMode(1);
	else	NES_SetCPUMode(0);
	
	CheckMenuRadioItem(hMenu,ID_DEBUG_LEVEL1,ID_DEBUG_LEVEL4,ID_DEBUG_LEVEL1,MF_BYCOMMAND);

	NES_UpdateInterface();
}

void	NES_SaveSettings (void)
{
	HKEY SettingsBase;
	if (RegOpenKeyEx(HKEY_CURRENT_USER,"SOFTWARE\\Nintendulator\\",0,KEY_ALL_ACCESS,&SettingsBase))
		RegCreateKeyEx(HKEY_CURRENT_USER,"SOFTWARE\\Nintendulator\\",0,"NintendulatorClass",REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&SettingsBase,NULL);
	RegSetValueEx(SettingsBase,"SoundEnabled",0,REG_DWORD,(unsigned char *)&NES.SoundEnabled,4);
	RegSetValueEx(SettingsBase,"SizeMult"    ,0,REG_DWORD,(unsigned char *)&SizeMult        ,4);
	RegSetValueEx(SettingsBase,"FSkip"       ,0,REG_DWORD,(unsigned char *)&GFX.FSkip       ,4);
	RegSetValueEx(SettingsBase,"aFSkip"      ,0,REG_DWORD,(unsigned char *)&GFX.aFSkip      ,4);
	RegSetValueEx(SettingsBase,"PPUMode"     ,0,REG_DWORD,(unsigned char *)&PPU.IsPAL       ,4);

	RegSetValueEx(SettingsBase,"Port1T"  ,0,REG_DWORD,(unsigned char *)&Controllers.Port1.Type  ,sizeof(Controllers.Port1.Type));
	RegSetValueEx(SettingsBase,"Port2T"  ,0,REG_DWORD,(unsigned char *)&Controllers.Port2.Type  ,sizeof(Controllers.Port2.Type));
	RegSetValueEx(SettingsBase,"FSPort1T",0,REG_DWORD,(unsigned char *)&Controllers.FSPort1.Type,sizeof(Controllers.FSPort1.Type));
	RegSetValueEx(SettingsBase,"FSPort2T",0,REG_DWORD,(unsigned char *)&Controllers.FSPort2.Type,sizeof(Controllers.FSPort2.Type));
	RegSetValueEx(SettingsBase,"FSPort3T",0,REG_DWORD,(unsigned char *)&Controllers.FSPort3.Type,sizeof(Controllers.FSPort3.Type));
	RegSetValueEx(SettingsBase,"FSPort4T",0,REG_DWORD,(unsigned char *)&Controllers.FSPort4.Type,sizeof(Controllers.FSPort4.Type));
	RegSetValueEx(SettingsBase,"ExpPortT",0,REG_DWORD,(unsigned char *)&Controllers.ExpPort.Type,sizeof(Controllers.ExpPort.Type));
	
	RegSetValueEx(SettingsBase,"Port1D"  ,0,REG_BINARY,(unsigned char *)Controllers.Port1.Buttons  ,sizeof(Controllers.Port1.Buttons));
	RegSetValueEx(SettingsBase,"Port2D"  ,0,REG_BINARY,(unsigned char *)Controllers.Port2.Buttons  ,sizeof(Controllers.Port2.Buttons));
	RegSetValueEx(SettingsBase,"FSPort1D",0,REG_BINARY,(unsigned char *)Controllers.FSPort1.Buttons,sizeof(Controllers.FSPort1.Buttons));
	RegSetValueEx(SettingsBase,"FSPort2D",0,REG_BINARY,(unsigned char *)Controllers.FSPort2.Buttons,sizeof(Controllers.FSPort2.Buttons));
	RegSetValueEx(SettingsBase,"FSPort3D",0,REG_BINARY,(unsigned char *)Controllers.FSPort3.Buttons,sizeof(Controllers.FSPort3.Buttons));
	RegSetValueEx(SettingsBase,"FSPort4D",0,REG_BINARY,(unsigned char *)Controllers.FSPort4.Buttons,sizeof(Controllers.FSPort4.Buttons));
	RegSetValueEx(SettingsBase,"ExpPortD",0,REG_BINARY,(unsigned char *)Controllers.ExpPort.Buttons,sizeof(Controllers.ExpPort.Buttons));
	
	RegCloseKey(SettingsBase);
}

void	NES_Repaint (void)
{
	GFX_Repaint();
}
